import os
from unittest import TestCase, mock

from b2luigi.core import utils


class ProductDictTestCase(TestCase):
    def test_basic_usage(self):
        kwargs = list(utils.product_dict(first_arg=[1, 2, 3], second_arg=["a", "b"]))

        self.assertEqual(len(kwargs), 6)

        self.assertIn({"first_arg": 1, "second_arg": "a"}, kwargs)
        self.assertIn({"first_arg": 1, "second_arg": "b"}, kwargs)
        self.assertIn({"first_arg": 2, "second_arg": "a"}, kwargs)
        self.assertIn({"first_arg": 2, "second_arg": "b"}, kwargs)
        self.assertIn({"first_arg": 3, "second_arg": "a"}, kwargs)
        self.assertIn({"first_arg": 3, "second_arg": "b"}, kwargs)

        kwargs = list(utils.product_dict(first_arg=[1, 2, 3]))

        self.assertEqual(len(kwargs), 3)

        self.assertIn({"first_arg": 1}, kwargs)
        self.assertIn({"first_arg": 2}, kwargs)
        self.assertIn({"first_arg": 3}, kwargs)

        kwargs = list(utils.product_dict(first_arg=[1, 2, 3], second_arg=[]))

        self.assertEqual(len(kwargs), 0)

        kwargs = list(utils.product_dict(first_arg=[1], second_arg=["a"]))

        self.assertEqual(len(kwargs), 1)

        self.assertIn({"first_arg": 1, "second_arg": "a"}, kwargs)


class FlattenTestCase(TestCase):
    def test_list_input(self):
        inputs = [{"key1": "value1"}, {"key2": "value2"}]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "value1")
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], "value2")

        inputs = [{"key1": "value1"}, {"key2": "value2"}, {"key1": "repeated"}]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "repeated")
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], "value2")

        inputs = []

        outputs = utils.flatten_to_dict(inputs)

        self.assertFalse(outputs)

        inputs = [{"key1": "value1"}, "value2"]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "value1")
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], "value2")

        inputs = ["value1", "value2"]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("value1", outputs)
        self.assertEqual(outputs["value1"], "value1")
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], "value2")

    def test_raw_input(self):
        inputs = {"key1": "value1"}

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "value1")

        inputs = "value1"

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("value1", outputs)
        self.assertEqual(outputs["value1"], "value1")

    def test_list_of_list_input(self):
        inputs = [{"key1": "value1"}, {"key2": "value2"}]

        outputs = utils.flatten_to_list_of_dicts(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1"])
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], ["value2"])

        inputs = [{"key1": "value1"}, {"key2": "value2"}, {"key1": "repeated"}]

        outputs = utils.flatten_to_list_of_dicts(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1", "repeated"])
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], ["value2"])

        inputs = []

        outputs = utils.flatten_to_list_of_dicts(inputs)

        self.assertFalse(outputs)

        inputs = [{"key1": "value1"}, "value2"]

        outputs = utils.flatten_to_list_of_dicts(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1"])
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], ["value2"])

        inputs = ["value1", "value2"]

        outputs = utils.flatten_to_list_of_dicts(inputs)

        self.assertIn("value1", outputs)
        self.assertEqual(outputs["value1"], ["value1"])
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], ["value2"])

        inputs = [[{"key1": "value1"}, {"key2": "value2"}], [{"key1": "repeated"}]]

        outputs = utils.flatten_to_list_of_dicts(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1", "repeated"])
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], ["value2"])


class MapFolderTestCase(TestCase):
    dummy_rel_dir = "./some/rel_dir"
    dummy_abs_dir = "/path/to/some/abs_dir"
    main_no_file_file_err_msg = "module '__main__' has no attribute '__file__'"

    def test_map_folder_abspath_identity(self):
        """Test that for an absolute path, map_folder returns and identity"""
        self.assertEqual(utils.map_folder(self.dummy_abs_dir), self.dummy_abs_dir)

    def test_map_folder_relpath(self):
        """
        Test map_folder with a relative input_folder, which joins it with ``__main__.__file__``
        """
        with mock.patch("__main__.__file__", self.dummy_abs_dir):
            mapped_folder = utils.map_folder(self.dummy_rel_dir)
            self.assertEqual(
                mapped_folder, os.path.join(self.dummy_abs_dir, mapped_folder)
            )

    def test_map_folder_abspath_identity_when_no_filename(self):
        """
        Test that for an absolute path, map_folder returns and identity even
        if ``get_filename`` would raise an ``AttributeError`` because ``__main__.__file__``
        is not available (e.g. in jupyter)
        """
        with mock.patch(
            "b2luigi.core.utils.get_filename",
            side_effect=AttributeError(self.main_no_file_file_err_msg),
        ):
            mapped_folder = utils.map_folder(self.dummy_abs_dir)
            self.assertEqual(mapped_folder, self.dummy_abs_dir)

    def test_map_folder_raises_attribute_error_for_relpath_when_no_filename(self):
        """
        Test that when ``get_filename`` returns an ``AttributeError`` b/c
        ``__main__.__file__`` is not available, ``map_folder`` also returns an
        ``AttributeError`` when the input folder is relative
        """
        with self.assertRaises(AttributeError):
            with mock.patch(
                "b2luigi.core.utils.get_filename",
                side_effect=AttributeError(self.main_no_file_file_err_msg),
            ):
                utils.map_folder(self.dummy_rel_dir)

    def _get_map_folder_error_mesage(self):
        """
        Get the error message that ``map_folder`` raises when ``get_filename``
        raises an attribute error because ``__main__.__file__`` is not
        accessible (e.g. in Jupyter)
        """
        try:
            with mock.patch(
                "b2luigi.core.utils.get_filename",
                side_effect=AttributeError(self.main_no_file_file_err_msg),
            ):
                utils.map_folder(self.dummy_rel_dir)
        except AttributeError as err:
            return str(err)
        raise RuntimeError("No AttributeError raised when calling ``utils.map_folder``")

    def test_original_message_in_error(self):
        """
        Check that the error message of ``map_folder`` still contains the
        original error message raised by ``get_filename`` during an
        ``AttributeError`` due to ``__main__.__file__`` not being accessible
        """
        message = self._get_map_folder_error_mesage()
        self.assertTrue(message.endswith(self.main_no_file_file_err_msg))

    def test_additional_info_added_to_error(self):
        """
        Check that the error message of ``map_folder`` adds additional
        information to the ``AttributeError`` raised by ``get_filename``
        """
        message = self._get_map_folder_error_mesage()
        self.assertTrue(message.startswith("Could not determine the current script location. "))



"""
Test helper functions for :py:class:`Gbasf2Process`.

Utilities that require a running gbasf2 or Dirac environment will not net
tested in this test case, only the functions that can be run independently.
"""

from b2luigi.batch.processes.htcondor import HTCondorProcess

from ..helpers import B2LuigiTestCase
from .batch_task_1 import MyTask
from unittest.mock import Mock
import os
import b2luigi


class TestHTCondorCreateSubmitFile(B2LuigiTestCase):

    def _get_htcondor_submit_file_string(self, task):
        # the _create_htcondor_submit_file is a method of the ``HTCondorProcess`` class, but it only uses its
        # class to obtain ``self.task``, to it's sufficient to provide a mock class for ``self``, `
        htcondor_mock_process = Mock()
        task.get_task_file_dir = lambda: self.test_dir
        htcondor_mock_process.task = task
        #  create submit file
        HTCondorProcess._create_htcondor_submit_file(htcondor_mock_process)
        # read submit file and return string
        submit_file_path = os.path.join(self.test_dir, "job.submit")
        with open(submit_file_path, "r") as submit_file:
            return submit_file.read()

    def test_minimal_submit_file(self):
        """
        Minimal  submit file should have expected shape:

            output = ...
            error = ..
            log = ..
            executable = executable_wrapper.sh
            queue 1
        """
        submit_file_lines = self._get_htcondor_submit_file_string(MyTask("some_parameter")).splitlines()
        self.assertIn("output = ", submit_file_lines[0])
        self.assertIn("error = ", submit_file_lines[1])
        self.assertIn("log = ", submit_file_lines[2])
        self.assertEqual("executable = executable_wrapper.sh", submit_file_lines[3])
        self.assertEqual("queue 1", submit_file_lines[4])

    def test_not_setting_job_name(self):
        submit_file_string = self._get_htcondor_submit_file_string(MyTask("some_parameter"))
        self.assertNotIn("JobBatchName", submit_file_string)

    def test_set_job_name_via_task_attribute(self):
        task = MyTask("some_parameter")
        task.job_name = "some_job_name"
        submit_file_lines = self._get_htcondor_submit_file_string(task).splitlines()
        self.assertIn("JobBatchName = some_job_name", submit_file_lines)

        b2luigi.set_setting("job_name", "some_job_name")
        submit_file_lines = self._get_htcondor_submit_file_string(MyTask("some_parameter")).splitlines()
        b2luigi.clear_setting("job_name")
        self.assertIn("JobBatchName = some_job_name", submit_file_lines)

    def test_set_job_name_is_overriden_by_htcondor_settings(self):
        """
        ``job_name`` is a global setting, but if the ``JobBatchName`` is set explicitly via the settings, we
        want that to override the global setting
        """
        task = MyTask("some_parameter")
        task.job_name = "job_name_global"
        htcondor_settings = {"JobBatchName": "job_name_htcondor"}
        b2luigi.set_setting("htcondor_settings", htcondor_settings)
        submit_file_lines = self._get_htcondor_submit_file_string(task).splitlines()
        b2luigi.clear_setting("htcondor_settings")
        self.assertNotIn("JobBatchName = job_name_global", submit_file_lines)
        self.assertIn("JobBatchName = job_name_htcondor", submit_file_lines)



"""
Test helper functions for :py:class:`Gbasf2Process`.

Utilities that require a running gbasf2 or Dirac environment will not net
tested in this test case, only the functions that can be run independently.
"""

import unittest

from b2luigi.batch.processes.gbasf2 import (_get_lfn_upto_reschedule_number, get_unique_lfns,
                                            lfn_follows_gb2v5_convention)

# first test utilities for working with logical file names on the grid


class TestLFNFollowsGbasf2V5Convention(unittest.TestCase):
    def test_newstyle_lfn_is_true(self):
        lfn = 'Upsilon4SBpcandee_00000_job181817516_03.root'
        self.assertTrue(lfn_follows_gb2v5_convention(lfn))

    def test_oldstyle_lfn_is_false(self):
        lfn = 'my_ntuple_0001.root'
        self.assertFalse(lfn_follows_gb2v5_convention(lfn))


class TestLFNUptoRescheduleNumber(unittest.TestCase):
    def test_strings_equal_camelcase_lfn(self):
        lfn = 'Upsilon4SBpcandee_00000_job181817516_03.root'
        self.assertEqual(_get_lfn_upto_reschedule_number(lfn), 'Upsilon4SBpcandee_00000_job181817516')

    def test_strings_equal_snakecase_lfn(self):
        lfn = 'my_ntuple_name_00000_job181817516_03.root'
        self.assertEqual(_get_lfn_upto_reschedule_number(lfn), 'my_ntuple_name_00000_job181817516')

    def test_oldstyle_lfn_raises_error(self):
        lfn = 'my_ntuple_0001.root'
        with self.assertRaises(ValueError):
            _get_lfn_upto_reschedule_number(lfn)


class TestGetUniqueLFNS(unittest.TestCase):
    def setUp(self):
        # list of logical path names on the grid that only differe by the reschedule number (last two digits)
        self.lfns_with_duplicates = [
            'Upsilon4SBpcandee_00000_job181817516_00.root',
            'Upsilon4SBpcandee_00001_job181817517_00.root',
            'Upsilon4SBpcandee_00002_job181817518_00.root',
            'Upsilon4SBpcandee_00003_job181817519_00.root',
            'Upsilon4SBpcandee_00004_job181817520_00.root',
            'Upsilon4SBpcandee_00005_job181817521_00.root',
            'Upsilon4SBpcandee_00006_job181817522_00.root',
            'Upsilon4SBpcandee_00007_job181817523_00.root',
            'Upsilon4SBpcandee_00008_job181817524_00.root',
            'Upsilon4SBpcandee_00009_job181817525_00.root',
            'Upsilon4SBpcandee_00010_job181817526_00.root',
            'Upsilon4SBpcandee_00011_job181817527_00.root',
            'Upsilon4SBpcandee_00012_job181817528_00.root',
            'Upsilon4SBpcandee_00012_job181817528_00.root',
            'Upsilon4SBpcandee_00012_job181817528_02.root',
            'Upsilon4SBpcandee_00013_job181817529_00.root',
            'Upsilon4SBpcandee_00014_job181817530_00.root',
            'Upsilon4SBpcandee_00015_job181817531_00.root',
            'Upsilon4SBpcandee_00016_job181817532_00.root',
            'Upsilon4SBpcandee_00017_job181817533_00.root',
            'Upsilon4SBpcandee_00017_job181817533_01.root',
            'Upsilon4SBpcandee_00017_job181817533_02.root',
            'Upsilon4SBpcandee_00018_job181817534_01.root',
            'Upsilon4SBpcandee_00019_job181817535_00.root',
        ]
        # create same input but with more underscores in initial root file name (snake case)
        # to test whether the string splitting logic is stable in that case
        self.lfns_with_duplicates_snake_case = [
            lfn.replace("Upsilon4SBpcandee", "u4s_Bp_cand_") for lfn in self.lfns_with_duplicates
        ]
        self.unique_lfns = {
            'Upsilon4SBpcandee_00000_job181817516_00.root',
            'Upsilon4SBpcandee_00001_job181817517_00.root',
            'Upsilon4SBpcandee_00002_job181817518_00.root',
            'Upsilon4SBpcandee_00003_job181817519_00.root',
            'Upsilon4SBpcandee_00004_job181817520_00.root',
            'Upsilon4SBpcandee_00005_job181817521_00.root',
            'Upsilon4SBpcandee_00006_job181817522_00.root',
            'Upsilon4SBpcandee_00007_job181817523_00.root',
            'Upsilon4SBpcandee_00008_job181817524_00.root',
            'Upsilon4SBpcandee_00009_job181817525_00.root',
            'Upsilon4SBpcandee_00010_job181817526_00.root',
            'Upsilon4SBpcandee_00011_job181817527_00.root',
            'Upsilon4SBpcandee_00012_job181817528_02.root',
            'Upsilon4SBpcandee_00013_job181817529_00.root',
            'Upsilon4SBpcandee_00014_job181817530_00.root',
            'Upsilon4SBpcandee_00015_job181817531_00.root',
            'Upsilon4SBpcandee_00016_job181817532_00.root',
            'Upsilon4SBpcandee_00017_job181817533_02.root',
            'Upsilon4SBpcandee_00018_job181817534_01.root',
            'Upsilon4SBpcandee_00019_job181817535_00.root',
        }
        self.unique_lfns_snake_case = {
            lfn.replace("Upsilon4SBpcandee", "u4s_Bp_cand_") for lfn in self.unique_lfns
        }

    def test_sets_equal(self):
        unique_lfns = get_unique_lfns(self.lfns_with_duplicates)
        self.assertSetEqual(unique_lfns, self.unique_lfns)

    def test_count_equal(self):
        unique_lfns = get_unique_lfns(self.lfns_with_duplicates)
        self.assertCountEqual(unique_lfns, self.unique_lfns)

    def test_sets_equal_underscore_in_file_name(self):
        unique_lfns_snake_case = get_unique_lfns(self.lfns_with_duplicates_snake_case)
        self.assertSetEqual(unique_lfns_snake_case, self.unique_lfns_snake_case)

    def test_sets_equal_input_as_set(self):
        unique_lfns = get_unique_lfns(set(self.lfns_with_duplicates))
        self.assertSetEqual(unique_lfns, self.unique_lfns)

    def test_sets_equal_underscore_in_file_name_input_as_set(self):
        unique_lfns_snake_case = get_unique_lfns(set(self.lfns_with_duplicates_snake_case))
        self.assertSetEqual(unique_lfns_snake_case, self.unique_lfns_snake_case)



from b2luigi.basf2_helper.targets import ROOTLocalTarget
from b2luigi.basf2_helper.tasks import (Basf2FileMergeTask, Basf2nTupleMergeTask, Basf2PathTask, Basf2Task, HaddTask,
                                        SimplifiedOutputBasf2Task)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['b2luigi',
 'b2luigi.basf2_helper',
 'b2luigi.batch',
 'b2luigi.batch.processes',
 'b2luigi.cli',
 'b2luigi.core']

package_data = \
{'': ['*'],
 'b2luigi.batch.processes': ['.mypy_cache/*',
                             '.mypy_cache/3.6/*',
                             '.mypy_cache/3.6/_typeshed/*',
                             '.mypy_cache/3.6/b2luigi/*',
                             '.mypy_cache/3.6/b2luigi/batch/*',
                             '.mypy_cache/3.6/b2luigi/cli/*',
                             '.mypy_cache/3.6/b2luigi/core/*',
                             '.mypy_cache/3.6/collections/*',
                             '.mypy_cache/3.6/importlib/*',
                             '.mypy_cache/3.6/json/*',
                             '.mypy_cache/3.6/os/*',
                             '.mypy_cache/3.8/*',
                             '.mypy_cache/3.8/_typeshed/*',
                             '.mypy_cache/3.8/b2luigi/*',
                             '.mypy_cache/3.8/b2luigi/batch/*',
                             '.mypy_cache/3.8/b2luigi/cli/*',
                             '.mypy_cache/3.8/b2luigi/core/*',
                             '.mypy_cache/3.8/collections/*',
                             '.mypy_cache/3.8/importlib/*',
                             '.mypy_cache/3.8/json/*',
                             '.mypy_cache/3.8/logging/*',
                             '.mypy_cache/3.8/os/*',
                             'gbasf2_utils/*',
                             'gbasf2_utils/.mypy_cache/*',
                             'gbasf2_utils/.mypy_cache/3.9/*',
                             'gbasf2_utils/.mypy_cache/3.9/_typeshed/*',
                             'gbasf2_utils/.mypy_cache/3.9/collections/*',
                             'gbasf2_utils/.mypy_cache/3.9/importlib/*',
                             'gbasf2_utils/.mypy_cache/3.9/os/*',
                             'templates/*'],
 'b2luigi.cli': ['.mypy_cache/*',
                 '.mypy_cache/3.9/*',
                 '.mypy_cache/3.9/_typeshed/*',
                 '.mypy_cache/3.9/asyncio/*',
                 '.mypy_cache/3.9/b2luigi/*',
                 '.mypy_cache/3.9/b2luigi/batch/*',
                 '.mypy_cache/3.9/b2luigi/cli/*',
                 '.mypy_cache/3.9/b2luigi/core/*',
                 '.mypy_cache/3.9/cachetools/*',
                 '.mypy_cache/3.9/collections/*',
                 '.mypy_cache/3.9/concurrent/*',
                 '.mypy_cache/3.9/concurrent/futures/*',
                 '.mypy_cache/3.9/ctypes/*',
                 '.mypy_cache/3.9/html/*',
                 '.mypy_cache/3.9/importlib/*',
                 '.mypy_cache/3.9/jinja2/*',
                 '.mypy_cache/3.9/json/*',
                 '.mypy_cache/3.9/logging/*',
                 '.mypy_cache/3.9/markupsafe/*',
                 '.mypy_cache/3.9/multiprocessing/*',
                 '.mypy_cache/3.9/os/*',
                 '.mypy_cache/3.9/urllib/*'],
 'b2luigi.core': ['.mypy_cache/*',
                  '.mypy_cache/3.8/*',
                  '.mypy_cache/3.8/_typeshed/*',
                  '.mypy_cache/3.8/b2luigi/*',
                  '.mypy_cache/3.8/b2luigi/batch/*',
                  '.mypy_cache/3.8/b2luigi/cli/*',
                  '.mypy_cache/3.8/b2luigi/core/*',
                  '.mypy_cache/3.8/collections/*',
                  '.mypy_cache/3.8/importlib/*',
                  '.mypy_cache/3.8/json/*',
                  '.mypy_cache/3.8/os/*']}

install_requires = \
['luigi>=3.0.2',
 'parse>=1.8.4',
 'GitPython>=2.1.11',
 'colorama>=0.3.9',
 'cachetools>=2.1.0',
 'jinja2']

setup(name='b2luigi',
      version='0.7.2',
      description='Task scheduling and batch running for basf2 jobs made simple',
      author='Nils Braun',
      author_email='nils.braun@kit.edu',
      url='https://github.com/nils-braun/b2luigi',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )
